import logging as log
import os
import re

from Configurables import (
    AddressKillerAlg,
    CaloClusterMCTruth,
    CaloDigit2MCLinks2Table,
    ChargedPP2MC,
    CopyLinePersistenceLocations,
    CopyParticle2PVRelationsFromLinePersistenceLocations,
    CopyProtoParticle2MCRelations,
    CopySignalMCParticles,
    DataOnDemandSvc,
    DataPacking__Pack_LHCb__CaloClusterPacker_ as PackCaloClusters,
    DataPacking__Unpack_LHCb__CaloClusterPacker_ as UnpackCaloClusters,
    DecodeRawEvent,
    DstConf,
    EventSelector,
    GaudiSequencer,
    HltPackedDataDecoder,
    HltRoutingBitsFilter,
    InputCopyStream,
    LHCbApp,
    LHCbConfigurableUser,
    LoKi__HDRFilter as HltDecReportsFilter,
    LumiAlgsConf,
    NeutralPP2MC,
    OutputStream,
    P2MCPFromProtoP,
    PackCaloHypo as PackCaloHypos,
    PackMCParticle,
    PackMCVertex,
    PackParticlesAndVertices,
    PrintDuplicates,
    RawEventFormatConf,
    RawEventJuggler,
    RecSummaryFromSelReports,
    RecombineRawEvent,
    RecordStream,
    TCKLinePersistenceSvc,
    TeslaLineChecker,
    TeslaReportAlgo,
    TrackAssociator,
    TrackSys,
    TurboConf,
    TurboPrescaler,
    UnpackCaloHypo as UnpackCaloHypos,
    UnpackParticlesAndVertices,
    bankKiller
)
from Gaudi.Configuration import *
from GaudiConf import IOExtension, IOHelper
from GaudiConf import PersistRecoConf
import GaudiKernel.ProcessJobOptions
from LHCbKernel.Configuration import *


class Tesla(LHCbConfigurableUser):
    __used_configurables__ = [ LHCbApp, LumiAlgsConf, RawEventJuggler, DecodeRawEvent, RawEventFormatConf, DstConf, RecombineRawEvent, PackParticlesAndVertices, TrackAssociator, ChargedPP2MC, TrackSys, TurboConf]

    __slots__ = {
            "EvtMax"		: -1    	# Maximum number of events to process
          , "SkipEvents"	: 0		# Skip to event
          , "Simulation"	: True 		# True implies use SimCond
          , "DataType"		: '2012' 	# Data type, can be [ 'DC06','2008' etc...]
          , "InputType"		: 'DST' 	# Input type, can be [ 'DST','LDST' etc...]
          , "DDDBtag" 		: 'default' 	# default as set in DDDBConf for DataType
          , "CondDBtag" 	: 'default' 	# default as set in DDDBConf for DataType
          , 'Persistency' 	: '' 		# None, Root or Pool?
          , 'OutputLevel' 	: 4 		# Er, output level
          , 'PackerOutputLevel' : 4             # Packer output level
          , 'DuplicateCheck' 	: False		# Do we want to add the algorithm to check for duplicates
          , 'KillInputTurbo' 	: False		# If rerunning Tesla, kill input Turbo locations
          , "outputFile" 	: 'Tesla.dst' 	# output filename
          , "outputPrefix"      : ''            # This is prepended to the stream name
          , "outputSuffix"      : '.dst'        # This is appended to the stream name
          , 'WriteFSR'    	: False 	# copy FSRs as required
          , 'PV'	        : "Offline"     # Associate to the PV chosen by the HLT or the offline one
          , 'VertRepLoc'	: "Hlt2"        # Do we use the maker from Hlt1 or Hlt2
          , 'TriggerLines'	: []            # Which trigger lines
          , 'Mode'	        : "Offline"     # "Online" (strip unnecessary banks and run lumi algorithms) or "Offline"?
                                            # "Online" means running on data from the pit
          , 'Pack'	        : True          # Do we want to pack the objects?
          , 'RawFormatVersion'	: 0.2           # Which banks form the Turbo stream
          , 'Monitors'	        : []            # Histogram monitors to run in the job
          , 'Histogram'  	: ""            # Name of histogram file
          , 'VetoJuggle'  	: True          # True if raw event removal not required in online mode
          , 'IgnoreDQFlags'  	: True          # Pass IgnoreDQFlags to LHCbApp
          , 'PersistRecoMC'  	: True          # Match persistReco objects to MC
          , 'ProtoTypesPR'  	: ["long","down"] # Which protoparticle types are in Turbo++
          , 'KillInputHlt2Reps'	: False         # Kill input Hlt2 reps to save space
          , 'InputHlt2RepsToKeep' : []          # List of selection names of reports to keep (if KillInputHlt2Reps is True)
          , 'HDRFilter'	        : False         # Filter on trigger lines or not
          , 'Park'	        : False         # If prescaled streams, make a parked version
         # Remove maybe
          , 'NonPrescaleList'   : []            # List of Turbo++ lines that are not prescaled
          , 'RemovePrescaleExcl': False         # Remove DST data banks in exclusively prescaled lines
          , 'PRtracks'  	: []            # PR track locations
          , 'PRchargedPP'  	: []            # PR protoparticle locations
          , 'PRclusters'  	: []            # PR cluster locations
          , 'PRneutralPP'  	: []            # PR neutral protoparticle locations
          , 'Streams'           : {}            # Streams, empty if no streaming
          , 'ValidateStreams'   : False         # Run validation to compare streams and TriggerLines
          , 'EnableLineChecker' : True          # Enable the comparison of lines in the DecReports and the input to TeslaReportsAlgo
          , 'IgnoredLines'      : [ ]   # Regexes for lines excluded from the comparison of DecReports and input to TeslaReportsAlgo
          , 'FilterMC' : False # Save a subset of the input MC particles and vertices under `Tesla.base, mimicking a microDST writer`
          , 'ILinePersistenceSvc' : "TCKLinePersistenceSvc"   # Implementation of the ILinePersistenceSvc to use (for >= 2017 data types)
          , 'CandidatesFromSelReports' : False  # If True and DataType >= 2017, take Turbo candidates from Hlt2SelReports rather than
                                                # from the DstData raw bank (i.e. behave as if DataType =< 2016)
          }
    _propertyDocDct ={
            "EvtMax"		: "Maximum number of events to process, default all"
            , "SkipEvents"	: "Events to skip, default     0"
            , "Simulation"	: "True implies use SimCond"
            , "DataType"	: "Data type, can be [ 'DC06','2008' ...]"
            , "InputType"	: "Input type, can be [ 'DST','LDST' ...]"
            , "DDDBtag" 	: "Databse tag, default as set in DDDBConf for DataType"
            , "CondDBtag" 	: "Databse tag, default as set in DDDBConf for DataType"
            , 'Persistency' 	: "Root or Pool?"
            , 'OutputLevel' 	: "Output level"
            , 'PackerOutputLevel': "Packer output level"
            , 'DuplicateCheck' 	: "Add the test for duplicates?"
            , 'KillInputTurbo' 	: "Are we rerunning Tesla"
            , "outputFile" 	: 'output filename, automatically selects MDF or InputCopyStream'
            , "outputPrefix"    : 'This is prepended to the stream name'
            , "outputSuffix"    : 'This is appended to the stream name'
            , 'WriteFSR'    	: 'copy FSRs as required'
            , 'PV'     	        : 'Associate to the PV chosen by the HLT or the offline one'
            , 'VertRepLoc'	: 'Do we use the maker from Hlt1 or Hlt2'
            , 'TriggerLines'    : 'Which trigger line to process'
            , 'Mode'     	: '"Online" (strip unnecessary banks and run lumi algorithms) or "Offline"?'
            , 'Pack'     	: 'Do we want to pack the object?'
            , 'RawFormatVersion': 'Which banks form the Turbo stream'
            , 'Monitors'        : 'Histogram monitors to run in the job'
            , 'Histogram'       : 'File name for histogram file'
            , 'VetoJuggle'      : 'Do we want to stop raw bank removal (assume happened further upstream)'
            , 'IgnoreDQFlags'  	: 'Pass IgnoreDQFlags to LHCbApp'
            , 'PersistRecoMC'  	: 'Match persistReco objects to MC'
            , 'ProtoTypesPR'  	: 'Which protoparticle types are in Turbo++'
            , 'KillInputHlt2Reps':'Kill input Hlt2 reps to save space'
            , 'InputHlt2RepsToKeep': 'List of selection names of reports to keep (if KillInputHlt2Reps is True)'
            , 'HDRFilter'	: 'Filter on trigger lines or not'
            , 'RemovePrescaleExcl': 'Remove DST data banks in exclusively prescaled lines'
            , 'Park'	        : 'If prescaled streams, make a parked version'
            , 'NonPrescaleList' : 'List of Turbo++ lines that are not prescaled'
            , 'PRtracks'  	: 'PR track locations'
            , 'PRchargedPP'  	: 'PR protoparticle locations'
            , 'PRclusters'  	: 'PR cluster locations'
            , 'PRneutralPP'  	: 'PR neutral protoparticle locations'
            , 'Streams'         : 'Streams, empty if no streaming'
            , 'ValidateStreams' : 'Run validation to compare streams and TriggerLines'
            , 'EnableLineChecker' : 'Enable the comparison of lines in the DecReports and the input to TeslaReportsAlgo'
            , 'IgnoredLines'      : 'Regexes for lines excluded from the comparison of DecReports and input to TeslaReportsAlgo'
            , 'FilterMC' : "Save a subset of the input MC particles and vertices under `Tesla.base`, mimicking a microDST writer"
            , 'ILinePersistenceSvc' : "Implementation of the ILinePersistenceSvc to use (for >= 2017 data types)"
            , 'CandidatesFromSelReports' : ("If True and DataType >= 2017, take Turbo candidates from Hlt2SelReports rather than "
                                            "from the DstData raw bank (i.e. behave as if DataType =< 2016)")

            }

    writerName = "DstWriter"
    teslaSeq = GaudiSequencer("TeslaSequence")
    base = "Turbo/"


    def _safeSet(self,conf,param):
        for p in param:
            if (not self.isPropertySet(p)) or conf.isPropertySet(p):
                continue
            conf.setProp(p,self.getProp(p))

    def _configureDataOnDemand(self) :
        if 'DataOnDemandSvc' in ApplicationMgr().ExtSvc :
            ApplicationMgr().ExtSvc.pop('DataOnDemandSvc')
        else:
            dod = DataOnDemandSvc()
        if dod not in ApplicationMgr().ExtSvc :
            ApplicationMgr().ExtSvc.append( dod )

    def _configureHistos(self):
        monitorSeq = GaudiSequencer('TelaMonitors')
        monitorSeq.IgnoreFilterPassed = True
        monitorSeq.Members += self.getProp('Monitors')
        self.teslaSeq.Members += [ monitorSeq ]
        #
        ApplicationMgr().HistogramPersistency = "ROOT"

        if ( self.getProp("Histogram") != "" ):
            HistogramPersistencySvc().OutputFile = self.getProp("Histogram")

    def _configureTrackTruth(self,assocpp,trackLoc) :
        assoctr = TrackAssociator("TurboAssocTr"+trackLoc)
        assoctr.OutputLevel = self.getProp('OutputLevel')
        assoctr.TracksInContainer = trackLoc
        assocpp.TrackLocations += [ trackLoc ]
        # Add it to a sequence
        return assoctr

    def _configureDigitsTruth(self) :
        assocdigits = CaloDigit2MCLinks2Table("TurboDigitAssoc")
        return

    def _configureClustersAndProtosTruth(self,digits,clusters,protos) :
        retSeq = GaudiSequencer("NeutralTruth")
        clusterTabLoc = self.base + "Relations/CaloClusters"

        assoccluster = CaloClusterMCTruth("TurboClusterAssoc")
        assoccluster.OutputLevel = self.getProp('OutputLevel')
        assoccluster.Input = digits
        assoccluster.Output = clusterTabLoc
        assoccluster.Clusters+=clusters

        # When filtering MC, the relations table cloners will copy the tables
        # to /Event/Turbo for us. Otherwise we create them there directly
        protoTabLoc = "Relations/Turbo/NeutralPP2MC"
        if not self.getProp("FilterMC"):
            protoTabLoc = os.path.join(self.base, protoTabLoc)
        assocneutral = NeutralPP2MC("TurboNeutralPP2MC")
        assocneutral.InputData += protos
        assocneutral.OutputLevel = self.getProp('OutputLevel')
        assocneutral.OutputTable = protoTabLoc
        assocneutral.MCCaloTable = clusterTabLoc

        retSeq.Members += [assoccluster,assocneutral]
        return protoTabLoc, retSeq

    def _unpackMC(self):
        DataOnDemandSvc().NodeMap['/Event/MC']   = 'DataObject'
        DataOnDemandSvc().AlgMap["MC/Particles"] = "UnpackMCParticle"
        DataOnDemandSvc().AlgMap["MC/Vertices"]  = "UnpackMCVertex"


    def _filterMCParticlesSequence(self, relations_locations):
        """Copy signal and associated MC particles to microDST-like locations.

        In the Stripping, the microDST machinery only saves a subset of the
        input MC particles and vertices. This subset contains two possibly
        overlapping contributions:

        1. The MCParticle objects in the signal decay tree; and
        2. The MCParticle objects which can be associated to the reconstructed
        objects selected by at least one stripping line.

        To mimic this behaviour in Tesla, we run almost the same set of
        microDST algorithms, using the relations tables Tesla makes to find the
        MC particles that have been associated to the reconstruction.
        """
        if not self.getProp('FilterMC'):
            return

        output_prefix = self.base.replace('/', '')

        # Algorithm to clone the existing relations tables and the MC particles
        # and vertices these tables point to from their original location to
        # the same location prefixed by /Event/Turbo
        copy_pp2mcp = CopyProtoParticle2MCRelations()
        copy_pp2mcp.InputLocations = relations_locations
        copy_pp2mcp.OutputPrefix = output_prefix
        # Don't use the assumed copy of the input ProtoParticle objects, as
        # Tesla doesn't need to copy them (they're already under /Event/Turbo)
        copy_pp2mcp.UseOriginalFrom = True

        # Algorithm to clone all MC particles and vertices that are associated
        # to the simulated signal process (using LHCb::MCParticle::fromSignal)
        copy_signal_mcp = CopySignalMCParticles()
        copy_signal_mcp.OutputPrefix = output_prefix
        # Don't copy the associated reconstruction, as Tesla already saves it
        copy_signal_mcp.SaveAssociatedRecoInfo = False

        algorithms = [copy_pp2mcp, copy_signal_mcp]
        seq = GaudiSequencer('TurboMCFiltering', Members=algorithms)

        return seq

    def _configureForOnline(self,stream):
        #
        DecodeRawEvent().DataOnDemand=False
        writer=InputCopyStream( self.writerName+stream )
        DstConf().setProp("SplitRawEventOutput", self.getProp("RawFormatVersion"))

        # we can only do this if we are setting one output writer
        if not self.getProp("Streams"):
            # Use RawEventJuggler to create the Turbo stream raw event format
            tck = "0x409f0045" # DUMMY
            TurboBanksSeq=GaudiSequencer("TurboBanksSeq")
            RawEventJuggler().TCK=tck
            RawEventJuggler().Input="Moore"
            RawEventJuggler().Output=self.getProp("RawFormatVersion")
            RawEventJuggler().Sequencer=TurboBanksSeq
            RawEventJuggler().WriterOptItemList=writer
            RawEventJuggler().KillExtraNodes=True
            RawEventJuggler().KillExtraBanks=True
            RawEventJuggler().KillExtraDirectories = True
            if not self.getProp('VetoJuggle'):
                self.teslaSeq.Members += [TurboBanksSeq]

        # we should already have configured the lumiAgls by now
        writer.AcceptAlgs += ["LumiSeq","PhysFilter"]

    def _configureLumi(self):

        # Begin Lumi configuration
        lumiSeq = GaudiSequencer("LumiSeq")
        #
        # Add ODIN decoder to LumiSeq ***
        from DAQSys.Decoders import DecoderDB
        CreateODIN=DecoderDB["createODIN"].setup()
        #********************************
        #
        # Main algorithm config
        lumiCounters = GaudiSequencer("LumiCounters")
        lumiCounters.Members+=[CreateODIN]
        lumiSeq.Members += [ lumiCounters ]

        LumiAlgsConf().LumiSequencer = lumiCounters
        LumiAlgsConf().OutputLevel = self.getProp('OutputLevel')
        LumiAlgsConf().InputType = "MDF"
        #
        # Filter out Lumi only triggers from further processing, but still write to output
        # Trigger masks changed in 2016, see LHCBPS-1486
        physFilterRequireMask = []
        if self.getProp( "DataType" ) in ["2012","2015"]: # 2012 needed for nightlies tests.
            physFilterRequireMask = [ 0x0, 0x4, 0x0 ]
        else:
            physFilterRequireMask = [ 0x0, 0x0, 0x80000000 ]
        physFilter = HltRoutingBitsFilter( "PhysFilter", RequireMask = physFilterRequireMask )

        FSRWriter = RecordStream( "FSROutputStreamDstWriter")
        FSRWriter.OutputLevel = INFO
        self.teslaSeq.Members += [lumiSeq, physFilter]

    def _configureOutput(self):

        TeslaReportAlgoSeq = GaudiSequencer('TeslaReportAlgoSeq')

        lines = self.getProp('TriggerLines')
        turbo_streams = self.getProp('Streams')

        self.teslaSeq.Members += [ TeslaReportAlgoSeq ]

        # Setting up online
        if self.getProp('Mode') is "Online":
            from DAQSys.Decoders import DecoderDB
            Hlt1VertexReportsDecoder=DecoderDB["HltVertexReportsDecoder/Hlt1VertexReportsDecoder"].setup()
            Hlt2VertexReportsDecoder=DecoderDB["HltVertexReportsDecoder/Hlt2VertexReportsDecoder"].setup()
            Hlt2SelReportsDecoder=DecoderDB["HltSelReportsDecoder/Hlt2SelReportsDecoder"].setup()
            Hlt2DecReportsDecoder=DecoderDB["HltDecReportsDecoder/Hlt2DecReportsDecoder"].setup()
            odin=DecoderDB["createODIN"].setup()
            TeslaReportAlgoSeq.Members += [ odin, Hlt1VertexReportsDecoder, Hlt2VertexReportsDecoder, Hlt2SelReportsDecoder, Hlt2DecReportsDecoder ]

        totalLines=[]
        if turbo_streams:
            lines_left_over = lines[:]
            missing_lines = []
            for stream in turbo_streams:
                for line in turbo_streams[stream]['lines']:
                    totalLines.append(line)
                    try:
                        lines_left_over.remove(line)
                    except ValueError:
                        missing_lines.append(line)
            if self.getProp('ValidateStreams'):
                msg = "Lines missing from streaming:\n  {}\nLines not in turbo but in streams:\n  {}\n".format(
                    '\n  '.join(lines_left_over), '\n  '.join(missing_lines))
                assert lines_left_over==[] and missing_lines==[], msg
                print("Turbo validated")
        else:
            totalLines = lines

        if self.getProp('EnableLineChecker'):
            lineChecker = TeslaLineChecker("TeslaLineChecker")
            lineChecker.RequestedLines = totalLines
            lineChecker.IgnoredLines = self.getProp("IgnoredLines")
            TeslaReportAlgoSeq.Members += [ lineChecker ]
        trig1 = self._configureReportAlg(totalLines)
        TeslaReportAlgoSeq.Members+=[trig1]

        # make use of PrintDuplicates algorithm
        dplic = PrintDuplicates("TurboDuplicates")
        dplic.OutputLevel = self.getProp('OutputLevel')
        for l in totalLines:
            dplic.Inputs+=[self.base + l + "/Particles"]
        if self.getProp('DuplicateCheck'):
            TeslaReportAlgoSeq.Members += [dplic]

        # Make protoparticle -> truth tables *****************************************
        inputType = self.getProp('InputType')
        if (inputType=='XDST') or (inputType=='LDST'):

            # Get Turbo++ locations
            PRtracks=self.getProp("PRtracks")
            PRclusters=self.getProp("PRclusters")
            PRprotos=self.getProp("PRchargedPP")
            PRprotosN=self.getProp("PRneutralPP")
            protoTypes = self.getProp("ProtoTypesPR")

            # CHARGED ################################################################
            ChargedProtoSeq = GaudiSequencer("ChargedTruthSequencer")

            # Standard Turbo locations
            protocont = self.base + "Protos"
            trackcont = self.base + "Tracks"

            assocpp=ChargedPP2MC("TurboProtoAssocPP")
            assocpp.OutputLevel = self.getProp('OutputLevel')
            assocpp.VetoEmpty=True
            # When filtering MC, the relations table cloners will copy the
            # tables to /Event/Turbo for us. Otherwise we create them there
            # directly
            tesROOT = "/Event/"
            if not self.getProp("FilterMC"):
                tesROOT = os.path.join(tesROOT, self.base)
            assocpp.RootInTES=tesROOT
            # Start with an empty input list, so we don't accidentally create
            # relations for non-Turbo/non-PersistReco objects
            assocpp.InputData = []

            # Add Turbo++ track and protoparticle locations
            truthSeqStandard = self._configureTrackTruth(assocpp,trackcont)
            ChargedProtoSeq.Members+=[truthSeqStandard]
            for PRtype in self.getProp("ProtoTypesPR"):
                matchedTracks = [s for s in PRtracks if PRtype in s.lower()]
                matchedProtos = [s for s in PRprotos if PRtype in s.lower()]
                if len(matchedTracks)>1 or len(matchedProtos)>1:
                    raise ValueError("Multiple instances of type "+PRtype+" have been detected")

                trackTruthSeq = self._configureTrackTruth(assocpp,matchedTracks[0])
                print("Adding additional PR protoparticles for truth matching: "+matchedProtos[0])
                assocpp.InputData += [ matchedProtos[0] ]
                ChargedProtoSeq.Members+=[trackTruthSeq]

            # Add container of merged long and downstream PersistReco PPs
            # TODO this should be in the list in the configuration file
            assocpp.InputData += ['Hlt2/Protos/Charged']
            print("Adding additional PR protoparticles for truth matching: "+assocpp.InputData[-1])
            trackTruthSeq = self._configureTrackTruth(assocpp, '/Event/Hlt2/TrackFitted/Charged')
            ChargedProtoSeq.Members += [trackTruthSeq]

            # Add standard Turbo ProtoParticles
            assocpp.InputData += [ protocont ]

            # Add final charged PP2MC algorithm
            ChargedProtoSeq.Members+=[assocpp]

            # NEUTRAL ################################################################
            NeutralProtoSeq = GaudiSequencer("NeutralTruthSequencer")

            ## Add the digits associator
            outputDigiLoc = tesROOT + "Digi2MCP"
            assocdigits = CaloDigit2MCLinks2Table("TurboDigitAssoc")
            assocdigits.OutputLevel = self.getProp('OutputLevel')
            outputiDigiLoc = tesROOT + "Relations/CaloDigits"
            assocdigits.Output = outputDigiLoc

            ## Add the cluster associator
            ## Finally configure the neutral PP2MC associator
            NeutralProtoSeq.Members+=[assocdigits]

            # Gather all protos and clusters
            neutralClustersTot=[]
            neutralProtosTot=[]

            # Add standard Turbo
            neutralClustersTot+=[self.base+"CaloClusters"]
            neutralProtosTot+=[self.base+"Protos"]

            # Add Turbo++
            for clusters in PRclusters:
                neutralClustersTot+=[clusters]
            for protos in PRprotosN:
                neutralProtosTot+=[protos]

            # Configure
            self._configureDigitsTruth()
            protoneutral, retSeq = self._configureClustersAndProtosTruth(outputDigiLoc,neutralClustersTot,neutralProtosTot)
            NeutralProtoSeq.Members+=[retSeq]

            # Add standard Turbo locations if not packing
            outputPPLoc = tesROOT+'Relations/Turbo/'
            if not self.getProp('Pack'):
                writer.OptItemList+=[
                        outputPPLoc + 'Protos' + '#99'
                        ]
            if not self.getProp('Pack'):
                writer.OptItemList+=[
                        protoneutral + '#99'
                        ]
            # Add final sequences
            TeslaReportAlgoSeq.Members+=[ChargedProtoSeq,NeutralProtoSeq]

            # Sequence to copy the relations tables, and the subset of MC
            # particles used by those tables, into /Event/Turbo
            if self.getProp("FilterMC"):
                relationsLocations = [os.path.join('Relations', path)
                                      for path in assocpp.InputData]
                relationsLocations.append(protoneutral)
                filterMCSeq = self._filterMCParticlesSequence(relationsLocations)
                TeslaReportAlgoSeq.Members += [filterMCSeq]

        if self.getProp('Pack'):
            packer = PackParticlesAndVertices( name = "TurboPacker",
                    InputStream        = "/Event"+"/"+self.base.rstrip('/'),
                    DeleteInput        = False,
                    EnableCheck        = False,
                    AlwaysCreateOutput = True)
            TeslaReportAlgoSeq.Members +=[packer]

            hypopacker = PackCaloHypos( name = "PackCaloHypos",
                    AlwaysCreateOutput = True,
                    DeleteInput        = False,
                    OutputLevel        = self.getProp('OutputLevel'),
                    InputName          = self.base+"CaloHypos",
                    OutputName         = self.base+"pRec/neutral/Hypos" )
            clusterpacker = PackCaloClusters( name = "PackCaloClusters",
                    AlwaysCreateOutput = True,
                    DeleteInput        = False,
                    OutputLevel        = self.getProp('OutputLevel'),
                    InputName          = self.base+"CaloClusters",
                    OutputName         = self.base+"pRec/neutral/Clusters" )
            TeslaReportAlgoSeq.Members +=[hypopacker,clusterpacker]

            # Pack the filtered MC particles and vertices
            if self.getProp('Simulation') and self.getProp('FilterMC'):
                mcp_packer = PackMCParticle(
                    'TurboPackMCParticle',
                    AlwaysCreateOutput=True,
                    DeleteInput=False,
                    InputName=self.base + 'MC/Particles',
                    OutputName=self.base + 'pSim/MCParticles'
                )
                mcv_packer = PackMCVertex(
                    'TurboPackMCVertex',
                    AlwaysCreateOutput=True,
                    DeleteInput=False,
                    InputName=self.base + 'MC/Vertices',
                    OutputName=self.base + 'pSim/MCVertices'
                )
                TeslaReportAlgoSeq.Members += [mcp_packer, mcv_packer]

            packer.OutputLevel = min(self.getProp('OutputLevel'),self.getProp('PackerOutputLevel'))

        TeslaOutputStreamsSequence = GaudiSequencer('TeslaOutputStreamsSequence')

        TeslaOutputStreamsSequence.ModeOR = True;
        TeslaOutputStreamsSequence.IgnoreFilterPassed = True

        if turbo_streams:
            if self.getProp('Park'):
                streamsOrig = turbo_streams
                for stream in list(streamsOrig.keys()):
                    ps = turbo_streams[stream].get('prescale', None)
                    if ps:
                        newDict={}
                        newDict['lines']=turbo_streams[stream]['lines']
                        pattern = re.compile("prescaled", re.IGNORECASE)
                        if "prescaled" in stream.lower():
                            turbo_streams[pattern.sub("parked", stream)]=newDict
                        else:
                            turbo_streams[stream+"parked"]=newDict
            for stream in turbo_streams:
                if self.getProp('Mode') is "Online":
                    self._configureForOnline(stream)
                TeslaOutputStreamsSequence.Members.append(self._configureOutputStream(stream, turbo_streams[stream]))
        else:
            if self.getProp('Mode') is "Online":
                self._configureForOnline('')
            TeslaOutputStreamsSequence.Members.append(self._configureOutputStream('', {'lines':lines}))

        # Add the output sequence
        TeslaReportAlgoSeq.Members += [TeslaOutputStreamsSequence]

    def _configureOutputStream(self, stream, lines_dict):
        '''
        The lines dictionary should have the following form:

        lines: The lines to include in the stream
        prescale: The amount to prescale, as a dictionary of lines. Uses the Prescale property if not present or None
        prescaleVersion: A version number
        park: The amount to park. A second line will be created, with the given amound and the label 'Park' appended (not implemented)

        Only the first key is required. The properties will be taken from self otherwise.

        Returns: a stream sequence
        '''


        # Building output name
        fname =""
        if not stream:
            fname = self.getProp("outputFile")
        else:
            fname = self.getProp("outputPrefix") + stream + self.getProp("outputSuffix")

        # Building sequencer with stream
        seq = GaudiSequencer('TeslaStreamSequence' + stream)

        # Required structure
        lines = lines_dict['lines']

        # Optional prescaling
        prescale = lines_dict.get('prescale', None)
        turboRepLoc="Turbo/DecReports"+stream
        if prescale:
            print "Requested to prescale stream: "+stream
            # if line not in dictionary, give it a prescale of 1
            for line in lines:
                if line not in prescale:
                    prescale[line]=1.0
            print prescale
            prescaleVersion = lines_dict.get('prescaleVersion', None)
            pscaler = TurboPrescaler("TurboPrescaler"+stream)
            pscaler.FilterOutput=True
            pscaler.ChosenOutputPrescales=prescale
            pscaler.PrescaleVersion=prescaleVersion
            pscaler.HltDecReportsLocation="Hlt2/DecReports"
            pscaler.ReportsOutputLoc=turboRepLoc
            seq.Members+=[pscaler]

            # Write new reports to the raw event
            #from Configurables import HltDecReportsWriter
            #decWriter = HltDecReportsWriter("Hlt2DecReportsWriter")
            #decWriter.SourceID=3
            #decWriter.InputHltDecReportsLocation=turboRepLoc
            #seq.Members+=[decWriter]

        #retrieve the persistency
        persistency=None
        if hasattr(self, "Persistency"):
            if self.getProp("Persistency") is not None:
                persistency=self.getProp("Persistency")

        iox=IOExtension(persistency)

        writer = InputCopyStream(self.writerName + stream)
        writer.OutputLevel = self.getProp('OutputLevel')

        writer.RequireAlgs += [seq]

        # If HDR requested, then filter on the output (required if streaming)
        if self.getProp("HDRFilter") or self.getProp("Streams"):
            filtCode='|'.join(["HLT_TURBOPASS_RE(\'"+l+"Decision\')" for l in lines])
            print("Filter code:")
            print(filtCode)
            HLTFilter = HltDecReportsFilter ('Hlt2_HLTFilter' + stream
                    , Code  = filtCode
                    , Location = "Hlt2/DecReports"
                    )
            writer.RequireAlgs+=[HLTFilter]
            seq.Members+=[HLTFilter]

        # Add our prescaled reports if present
        if prescale:
            writer.OptItemList+=[ turboRepLoc + "#99"]

        for l in lines:
            # IF NOT PACKING NEED TO SET THE LOCATIONS IN THE REPORT ALGORITHM
            if not self.getProp('Pack'):
                writer.OptItemList+=[ self.base + l + "/Particles#99"
                        , self.base + l + "/Vertices#99"
                        , self.base + l + "/_ReFitPVs#99"
                        , self.base + l + "/Particle2VertexRelations#99"
                        ]

        if not self.getProp('Pack'):
            writer.OptItemList+=[
                self.base + "Protos#99"
                ,self.base + "Tracks#99"
                ,self.base + "RichPIDs#99"
                ,self.base + "MuonPIDs#99"
                ,self.base + "CaloHypos#99"
                ,self.base + "CaloClusters#99"
                ]


        # Does this need to be in streaming? HENRYIII
        # Add shared places to writer
        # SB: Might as well support non-packing case.
        if not self.getProp('Pack'):
            writer.OptItemList+=[
                    self.base + "Primary#99"
                    , self.base + "Rec/Summary#99"
                    ]

        if self.getProp('Pack'):
            writer.OptItemList+=[
                    self.base+"pPhys/Particles#99"
                    ,self.base+"pPhys/Vertices#99"
                    ,self.base+"pPhys/RecVertices#99"
                    ,self.base+"pPhys/Relations#99"
                    ,self.base+"pPhys/PP2MCPRelations#99"
                    ,self.base+"pPhys/PartToRelatedInfoRelations#99"
                    ,self.base+"pRec/Track/Custom#99"
                    ,self.base+"pRec/Muon/CustomPIDs#99"
                    ,self.base+"pRec/Rich/CustomPIDs#99"
                    ,self.base+"pRec/ProtoP/Custom#99"
                    ,self.base+"pRec/neutral/Hypos#99"
                    ,self.base+"pRec/neutral/Clusters#99"
                    ,self.base+"Rec/Summary#99"
                    ,self.base+"pSim/MCParticles#99"
                    ,self.base+"pSim/MCVertices#99"
                    ]

        IOHelper(persistency,persistency).outStream(fname,writer,writeFSR=self.getProp('WriteFSR'))
        return seq

    def _configureReportAlg(self, lines):
        trig1 = TeslaReportAlgo("TeslaReportAlgo")
        trig1.OutputPrefix=self.base
        trig1.PV=self.getProp('PV')
        if self.getProp('VertRepLoc')=="Hlt1":
            trig1.VertRepLoc="Hlt1/VertexReports"
        else:
            trig1.VertRepLoc="Hlt2/VertexReports"
        trig1.PVLoc=self.base+"Primary"
        trig1.TriggerLines=lines
        trig1.OutputLevel=self.getProp('OutputLevel')
        return trig1

    def _configureOutputTurboSP(self):
        trigger_lines = self.getProp('TriggerLines')
        streams = self.getProp('Streams')
        simulation = self.getProp('Simulation')
        datatype = self.getProp('DataType')
        mode = self.getProp('Mode')
        online = mode == 'Online'
        raw_format = self.getProp('RawFormatVersion')
        vertex_report_location = self.getProp('VertRepLoc')
        enable_line_checker = self.getProp('EnableLineChecker')

        ok = (trigger_lines and not streams) or (streams and not trigger_lines)
        assert ok, ('Must define at least one of Tesla().Streams and '
                    'Tesla().TriggerLines, but not both')

        # Treat the no-streaming case as a single anonymous stream
        if not streams:
            streams = {'': {'lines': trigger_lines}}

        decoders_seq = GaudiSequencer('TeslaDecoders')
        if online:
            decoder_names = [
                'HltVertexReportsDecoder/Hlt2VertexReportsDecoder',
                'HltSelReportsDecoder/Hlt2SelReportsDecoder',
                'HltDecReportsDecoder/Hlt2DecReportsDecoder',
                'createODIN'
            ]
            if vertex_report_location == 'Hlt1':
                decoder_names.insert(0, 'HltVertexReportsDecoder/Hlt1VertexReportsDecoder')
            for name in decoder_names:
                from DAQSys.Decoders import DecoderDB
                decoder = DecoderDB[name].setup()
                decoders_seq.Members.append(decoder)

            DecodeRawEvent().DataOnDemand = False
            DstConf().setProp('SplitRawEventOutput', raw_format)

        # This decoder is setup by TurboConf
        packed_data_decoder = HltPackedDataDecoder('Hlt2PackedDataDecoder')
        decoders_seq.Members.append(packed_data_decoder)

        # Decode the RecSummary from the SelReports to /Event/Rec/Summary
        recsummary_decoder = RecSummaryFromSelReports()
        recsummary_decoder.InputHltSelReportsLocation = 'Hlt2/SelReports'
        decoders_seq.Members.append(recsummary_decoder)

        prpacking = PersistRecoConf.PersistRecoPacking(datatype)
        prunpackers = prpacking.unpackers()

        unpackers_seq = GaudiSequencer('TeslaUnpackers')
        # Need to run the packers in reverse order, due to object dependencies
        unpackers_seq.Members += prunpackers[::-1]
        unpack_psandvs = UnpackParticlesAndVertices(
            InputStream='/Event/Turbo'
        )
        unpackers_seq.Members.append(unpack_psandvs)

        stream_sequences = []
        for stream_name in streams:
            stream_seq = self._configureOutputTurboSPStream(
                stream_name,
                streams[stream_name]['lines'],
                prpacking
            )
            stream_sequences.append(stream_seq)

        # Kill any links the output file might have had to the input file
        address_killer = AddressKillerAlg()

        if enable_line_checker:
            all_lines = sum([d['lines'] for d in streams.values()], [])
            ignored_lines = self.getProp('IgnoredLines')
            line_checker = [TeslaLineChecker(
                RequestedLines=all_lines,
                IgnoredLines=ignored_lines
            )]
        else:
            line_checker = []

        # Kill the input DstData bank
        dstdata_killer = bankKiller(
            "Hlt2DstDataKiller",
            BankTypes=["DstData"]
        )

        streaming_seq = GaudiSequencer(
            'TeslaStreamsSequence',
            Members=(
                [decoders_seq, unpackers_seq] +
                line_checker +
                stream_sequences +
                [address_killer] +
                [dstdata_killer]
            ),
            IgnoreFilterPassed=True
        )

        return streaming_seq

    def _configureOutputTurboSPStream(self, name, lines, packing):
        """Return a sequence for streaming the lines into the named location.

        Copy line outputs for this stream to a stream-specific location.

        Keyword arguments:
        name -- Name of the stream
        lines -- List of HLT2 line names that belong in the stream
        packing -- Instance of PersistRecoConf.PersistRecoPacking
        """
        output_prefix = name.title()
        decisions = [l + 'Decision' for l in lines]
        # Naming convention for stream-specific algorithms
        namer = lambda x: '{0}ForStream{1}'.format(x, output_prefix)
        pack = self.getProp('Pack')
        write_fsr = self.getProp('WriteFSR')

        stream_seq = GaudiSequencer(namer('TeslaStreamSequence'))

        # Run the stream sequence iff at least one line in the stream fired
        filter_code = '|'.join([
            "HLT_TURBOPASS_RE('{0}')".format(l) for l in decisions
        ])
        filter = HltDecReportsFilter(
            namer('Hlt2_HLTFilter'),
            Code=filter_code
        )
        # Don't filter an anonymous stream by default
        if output_prefix or self.getProp('HDRFilter'):
            stream_seq.Members.append(filter)

        # No need to clone if the output prefix is empty (everything stays
        # under /Event/Turbo)
        if output_prefix:
            persistence_svc = self.getProp('ILinePersistenceSvc')
            ApplicationMgr().ExtSvc.append(persistence_svc)
            # Tell the persistence service to map any packed locations it knows
            # about to unpacked locations, as it is the unpacked locations we
            # want to copy to stream-specific locations
            TCKLinePersistenceSvc().ContainerMap = packing.packedToOutputLocationMap()
            container_cloner = CopyLinePersistenceLocations(
                namer('CopyLinePersistenceLocations'),
                OutputPrefix=output_prefix,
                LinesToCopy=decisions,
                ILinePersistenceSvc=persistence_svc
            )
            p2pv_cloner = CopyParticle2PVRelationsFromLinePersistenceLocations(
                namer('CopyP2PVRelationsFromLinePersistenceLocations'),
                OutputPrefix=output_prefix,
                LinesToCopy=decisions,
                ILinePersistenceSvc=persistence_svc
            )
            copy_line_outputs_seq = GaudiSequencer(
                namer('TeslaCopyLineOutputsSequence'),
                Members=[
                    container_cloner,
                    p2pv_cloner
                ],
                IgnoreFilterPassed=True
            )
            stream_seq.Members.append(copy_line_outputs_seq)

        tes_root = '/Event'
        # /Event/<stream name>
        stream_base = os.path.join(tes_root, output_prefix).rstrip('/')
        # /Event/<stream name>/Turbo
        turbo_base = os.path.join(stream_base, 'Turbo')
        required_output_locations = [
            os.path.join(tes_root, 'DAQ/ODIN#1'),
            os.path.join(tes_root, 'DAQ/RawEvent#1'),
            os.path.join(tes_root, 'Rec/Summary#1')
        ]
        optional_output_locations = []
        if pack:
            datatype = self.getProp('DataType')
            packers = []

            # Pack the PersistReco locations for this stream
            prpacking_inputs = {
                k: v.replace(tes_root, stream_base)
                for k, v in packing.outputs.items()
            }
            prpacking_descriptors = {
                datatype: {
                    k: d.copy(location=d.location.replace(tes_root, turbo_base))
                    for k, d in PersistRecoConf.standardDescriptors[datatype].items()
                }
            }
            prpacking_stream = PersistRecoConf.PersistRecoPacking(
                datatype,
                inputs=prpacking_inputs,
                descriptors=prpacking_descriptors
            )
            packers += prpacking_stream.packers(identifier=namer(''))

            # Pack everything else that we've cloned
            # (Packed locations made by PackParticlesAndVertices already exist
            # at the anonymous stream location)
            if output_prefix:
                psandvs_packer = PackParticlesAndVertices(
                    namer('TurboPacker'),
                    InputStream=turbo_base,
                    VetoedContainers=prpacking_stream.inputs.values()
                )
                packers.append(psandvs_packer)

            for packer in packers:
                packer.OutputLevel = min(self.getProp('OutputLevel'),
                                         self.getProp('PackerOutputLevel'))

            packing_seq = GaudiSequencer(
                namer('TeslaPackingSequence'),
                Members=packers,
                IgnoreFilterPassed=True
            )
            stream_seq.Members.append(packing_seq)

            optional_output_locations += [
                os.path.join(turbo_base, 'pPhys#*'),
                # XXX can we better take advantage of the automagic DaVinci
                # unpackers if we packed PersistReco stuff to just pRec?
                os.path.join(turbo_base, 'Hlt2/pRec#*'),
                # os.path.join(turbo_base, 'pRec#*')
            ]
        else:
            # Save everything under the stream prefix
            optional_output_locations += [
                os.path.join(turbo_base, '#*')
            ]

        if output_prefix:
            fname = '{0}{1}{2}'.format(
                self.getProp('outputPrefix'),
                output_prefix,
                self.getProp('outputSuffix')
            )
        else:
            fname = self.getProp('outputFile')

        writer = OutputStream(namer(self.writerName))
        writer.ItemList = required_output_locations
        writer.OptItemList = optional_output_locations
        # We should already have configured the lumiAlgs by now
        writer.AcceptAlgs = ['LumiSeq', 'PhysFilter']
        writer.RequireAlgs = [stream_seq]
        iohelper = IOHelper()
        iohelper.outStream(fname, writer, writeFSR=write_fsr)

        return stream_seq

    def _configureHlt2SelReportsKill(self):
        from Configurables import bankKiller, HltSelReportsStripper

        inputHlt2RepsToKeep = self.getProp('InputHlt2RepsToKeep')

        kill_seq = GaudiSequencer('Hlt2SelReportsKillerSeq')

        killer = bankKiller("Hlt2SelReportsBankKiller")
        killer.BankTypes = ["HltSelReports"]
        killer.DefaultIsKill = False
        killer.KillSourceID = 2 << 13
        kill_seq.Members.append(killer)

        # if the list of reports to keep is empty, do nothing more
        if inputHlt2RepsToKeep:
            assert all(name.endswith('Decision') for name in inputHlt2RepsToKeep)
            # By the time the stripper runs Hlt2/SelReports should exist
            stripper = HltSelReportsStripper('Hlt2SelReportsStripper')
            stripper.OutputHltSelReportsLocation = 'Hlt2/SelReportsFiltered'
            stripper.OutputHltObjectSummariesLocation = (
                str(stripper.OutputHltSelReportsLocation) + '/Candidates')
            stripper.SelectionNames = inputHlt2RepsToKeep
            stripper.OutputLevel = self.getProp('OutputLevel')
            kill_seq.Members.append(stripper)

            writer = HltSelReportsWriter('Hlt2SelReportsFilteredWriter')
            writer.SourceID = 2
            writer.InputHltSelReportsLocation = str(stripper.OutputHltSelReportsLocation)
            writer.OutputLevel = self.getProp('OutputLevel')
            kill_seq.Members.append(writer)

        return kill_seq

    def __apply_configuration__(self):
        ############## Set other properties ###########
        self._safeSet( LHCbApp(), ['EvtMax','SkipEvents','Simulation', 'DataType' , 'CondDBtag','DDDBtag'] )
        self._safeSet( TrackSys(), ['DataType'] )
        self._safeSet( TurboConf(), ['DataType'] )

        # Make sure DQFlags are ignored for Tesla productions
        if self.getProp('IgnoreDQFlags'):
            if not LHCbApp().isPropertySet( "IgnoreDQFlags" ) :
                LHCbApp().setProp( "IgnoreDQFlags", True )

        ApplicationMgr().AppName="Tesla, utilising DaVinci"
        #
        if self.getProp('Mode') is "Online":
            self.setProp('WriteFSR',True)
            self._configureLumi()
        else:
            DecodeRawEvent().DataOnDemand=True
            RecombineRawEvent()
            if self.getProp('Simulation')==True:
                self._unpackMC()
                TurboConf().setProp("PersistReco",True)
                # We don't want the PersistReco reconstruction under
                # /Event/Turbo, else it will be packed
                TurboConf().setProp("RootInTES", "/Event")

        if self.getProp('DataType') == '2017' and not self.getProp('CandidatesFromSelReports'):
            # Unpack the DstData raw bank
            TurboConf().setProp("RunPackedDataDecoder", True)
            streaming_seq = self._configureOutputTurboSP()
            self.teslaSeq.Members.append(streaming_seq)
        else:
            self._configureOutput()
        #
        EventSelector().PrintFreq = 1000

        # Add monitors if they are there
        if len(self.getProp('Monitors'))>0:
            self._configureHistos()

        if self.getProp('KillInputTurbo'):
            enk = EventNodeKiller('KillTurbo')
            enk.Nodes = [ "Turbo" ]
            ApplicationMgr().TopAlg.insert( 0,  enk.getFullName() )

        if self.getProp('KillInputHlt2Reps'):
            kill_selreports = self._configureHlt2SelReportsKill()
            self.teslaSeq.Members += [kill_selreports]

        ApplicationMgr().TopAlg+=[self.teslaSeq]
